# -*- coding: utf-8 -*-
import os

import crossplane
from .. import compare_parsed_and_built, here as tests_dir


def test_lex_lua_block_simple():
    dirname = os.path.join(tests_dir, 'configs', 'lua-block-simple')
    config = os.path.join(dirname, 'nginx.conf')
    tokens = crossplane.lex(config)
    assert list((token, line) for token, line, quoted in tokens) == [
        ('http', 1),
        ('{', 1),
        ('init_by_lua_block', 2),
        ('\n        print("Lua block code with curly brace str {")\n    ', 4),
        (';', 4),
        ('init_worker_by_lua_block', 5),
        ('\n        print("Work that every worker")\n    ', 7),
        (';', 7),
        ('body_filter_by_lua_block', 8),
        ('\n        local data, eof = ngx.arg[1], ngx.arg[2]\n    ', 10),
        (';', 10),
        ('header_filter_by_lua_block', 11),
        ('\n        ngx.header["content-length"] = nil\n    ', 13),
        (';', 13),
        ('server', 14),
        ('{', 14),
        ('listen', 15),
        ('127.0.0.1:8080', 15),
        (';', 15),
        ('location', 16),
        ('/', 16),
        ('{', 16),
        ('content_by_lua_block', 17),
        ('\n                ngx.say("I need no extra escaping here, for example: \\r\\nblah")\n            ', 19),
        (';', 19),
        ('return', 20),
        ('200', 20),
        ('foo bar baz', 20),
        (';', 20),
        ('}', 21),
        ('ssl_certificate_by_lua_block', 22),
        ('\n            print("About to initiate a new SSL handshake!")\n        ', 24),
        (';', 24),
        ('location', 25),
        ('/a', 25),
        ('{', 25),
        ('client_max_body_size', 26),
        ('100k', 26),
        (';', 26),
        ('client_body_buffer_size', 27),
        ('100k', 27),
        (';', 27),
        ('}', 28),
        ('}', 29),
        ('upstream', 31),
        ('foo', 31),
        ('{', 31),
        ('server', 32),
        ('127.0.0.1', 32),
        (';', 32),
        ('balancer_by_lua_block', 33),
        ('\n            -- use Lua to do something interesting here\n        ', 35),
        (';', 35),
        ('log_by_lua_block', 36),
        ('\n            print("I need no extra escaping here, for example: \\r\\nblah")\n        ', 38),
        (';', 38),
        ('}', 39),
        ('}', 40)
    ]


def test_lex_lua_block_larger():
    dirname = os.path.join(tests_dir, 'configs', 'lua-block-larger')
    config = os.path.join(dirname, 'nginx.conf')
    tokens = crossplane.lex(config)
    assert list((token, line) for token, line, quoted in tokens) == [
        ('http', 1),
        ('{', 1),
        ('content_by_lua_block', 2),
        (
            '\n        ngx.req.read_body()  -- explicitly read the req body'
            '\n        local data = ngx.req.get_body_data()'
            '\n        if data then'
            '\n            ngx.say("body data:")'
            '\n            ngx.print(data)'
            '\n            return'
            '\n        end'
            '\n'
            '\n        -- body may get buffered in a temp file:'
            '\n        local file = ngx.req.get_body_file()'
            '\n        if file then'
            '\n            ngx.say("body is in file ", file)'
            '\n        else'
            '\n            ngx.say("no body found")'
            '\n        end'
            '\n    ',
            18
        ),
        (';', 18),
        ('access_by_lua_block', 19),
        (
            '\n        -- check the client IP address is in our black list'
            '\n        if ngx.var.remote_addr == "132.5.72.3" then'
            '\n            ngx.exit(ngx.HTTP_FORBIDDEN)'
            '\n        end'
            '\n'
            '\n        -- check if the URI contains bad words'
            '\n        if ngx.var.uri and'
            '\n               string.match(ngx.var.request_body, "evil")'
            '\n        then'
            '\n            return ngx.redirect("/terms_of_use.html")'
            '\n        end'
            '\n'
            '\n        -- tests passed'
            '\n    ',
            33
        ),
        (';', 33),
        ('}', 34)
    ]


def test_lex_lua_block_tricky():
    dirname = os.path.join(tests_dir, 'configs', 'lua-block-tricky')
    config = os.path.join(dirname, 'nginx.conf')
    tokens = crossplane.lex(config)
    assert list((token, line) for token, line, quoted in tokens) == [
        ('http', 1),
        ('{', 1),
        ('server', 2),
        ('{', 2),
        ('listen', 3),
        ('127.0.0.1:8080', 3),
        (';', 3),
        ('server_name', 4),
        ('content_by_lua_block', 4),
        (';', 4),
        ("# make sure this doesn't trip up lexers", 4),
        ('set_by_lua_block', 5),
        ('$res', 5),
        (
            ' -- irregular lua block directive'
            '\n            local a = 32'
            '\n            local b = 56'
            '\n'
            '\n            ngx.var.diff = a - b;  -- write to $diff directly'
            '\n            return a + b;          -- return the $sum value normally'
            '\n        ',
            11
        ),
        (';', 11),
        ('rewrite_by_lua_block', 12),
        (
            ' -- have valid braces in Lua code and quotes around directive'
            '\n            do_something("hello, world!\\nhiya\\n")'
            '\n            a = { 1, 2, 3 }'
            '\n            btn = iup.button({title="ok"})'
            '\n        ',
            16
        ),
        (';', 16),
        ('}', 17),
        ('upstream', 18),
        ('content_by_lua_block', 18),
        ('{', 18),
        ('# stuff', 19),
        ('}', 20),
        ('}', 21)
    ]


def test_parse_lua_block_simple():
    dirname = os.path.join(tests_dir, 'configs', 'lua-block-simple')
    config = os.path.join(dirname, 'nginx.conf')
    payload = crossplane.parse(config)
    assert payload == {
        'status': 'ok',
        'errors': [],
        'config': [
            {
                'file': os.path.join(dirname, 'nginx.conf'),
                'status': 'ok',
                'errors': [],
                'parsed': [
                    {
                        'line': 1,
                        'args': [],
                        'block': [
                            {
                                'line': 2,
                                'args': ['\n        print("Lua block code with curly brace str {")\n    '],
                                'directive': 'init_by_lua_block'
                            },
                            {
                                'line': 5,
                                'args': ['\n        print("Work that every worker")\n    '],
                                'directive': 'init_worker_by_lua_block'
                            },
                            {
                                'line': 8,
                                'args': ['\n        local data, eof = ngx.arg[1], ngx.arg[2]\n    '],
                                'directive': 'body_filter_by_lua_block'
                            },
                            {
                                'line': 11,
                                'args': ['\n        ngx.header["content-length"] = nil\n    '],
                                'directive': 'header_filter_by_lua_block'
                            },
                            {
                                'line': 14,
                                'args': [],
                                'block': [
                                    {
                                        'line': 15,
                                        'args': ['127.0.0.1:8080'],
                                        'directive': 'listen'
                                    },
                                    {
                                        'line': 16,
                                        'args': ['/'],
                                        'block': [
                                            {
                                                'line': 17,
                                                'args': [
                                                    '\n                ngx.say("I need no extra escaping here, for example: \\r\\nblah")'
                                                    '\n            '
                                                ],
                                                'directive': 'content_by_lua_block'
                                            },
                                            {
                                                'line': 20,
                                                'args': ['200', 'foo bar baz'],
                                                'directive': 'return'
                                            }
                                        ],
                                        'directive': 'location'
                                    },
                                    {
                                        'line': 22,
                                        'args': [
                                            '\n            print("About to initiate a new SSL handshake!")'
                                            '\n        '
                                        ],
                                        'directive': 'ssl_certificate_by_lua_block'
                                    },
                                    {
                                        'line': 25,
                                        'args': ['/a'],
                                        'block': [
                                            {
                                                'line': 26,
                                                'args': ['100k'],
                                                'directive': 'client_max_body_size'
                                            },
                                            {
                                                'line': 27,
                                                'args': ['100k'],
                                                'directive': 'client_body_buffer_size'
                                            }
                                        ],
                                        'directive': 'location'
                                    }
                                ],
                                'directive': 'server'
                            },
                            {
                                'line': 31,
                                'args': ['foo'],
                                'block': [
                                    {
                                        'line': 32,
                                        'args': ['127.0.0.1'],
                                        'directive': 'server'
                                    },
                                    {
                                        'line': 33,
                                        'args': [
                                            '\n            -- use Lua to do something interesting here'
                                            '\n        '
                                        ],
                                        'directive': 'balancer_by_lua_block'
                                    },
                                    {
                                        'line': 36,
                                        'args': [
                                            '\n            print("I need no extra escaping here, for example: \\r\\nblah")'
                                            '\n        '
                                        ],
                                        'directive': 'log_by_lua_block'
                                    }
                                ],
                                'directive': 'upstream'
                            }
                        ],
                        'directive': 'http'
                    }
                ]
            }
        ]
    }


def test_parse_lua_block_tricky():
    dirname = os.path.join(tests_dir, 'configs', 'lua-block-tricky')
    config = os.path.join(dirname, 'nginx.conf')
    payload = crossplane.parse(config, comments=True)
    assert payload == {
        'status': 'ok',
        'errors': [],
        'config': [
            {
                'file': os.path.join(dirname, 'nginx.conf'),
                'status': 'ok',
                'errors': [],
                'parsed': [
                    {
                        'line': 1,
                        'args': [],
                        'block': [
                            {
                                'line': 2,
                                'args': [],
                                'block': [
                                    {
                                        'line': 3,
                                        'args': ['127.0.0.1:8080'],
                                        'directive': 'listen'
                                    },
                                    {
                                        'line': 4,
                                        'args': ['content_by_lua_block'],
                                        'directive': 'server_name'
                                    },
                                    {
                                        'comment': u" make sure this doesn't trip up lexers",
                                        'line': 4,
                                        'args': [],
                                        'directive': '#'
                                    },
                                    {
                                        'line': 5,
                                        'args': [
                                            '$res',
                                            ' -- irregular lua block directive'
                                            '\n            local a = 32'
                                            '\n            local b = 56'
                                            '\n'
                                            '\n            ngx.var.diff = a - b;  -- write to $diff directly'
                                            '\n            return a + b;          -- return the $sum value normally'
                                            '\n        '
                                        ],
                                        'directive': 'set_by_lua_block'
                                    },
                                    {
                                        'line': 12,
                                        'args': [
                                            ' -- have valid braces in Lua code and quotes around directive'
                                            '\n            do_something("hello, world!\\nhiya\\n")'
                                            '\n            a = { 1, 2, 3 }'
                                            '\n            btn = iup.button({title="ok"})'
                                            '\n        '
                                        ],
                                        'directive': 'rewrite_by_lua_block'
                                    }
                                ],
                                'directive': 'server'
                            },
                            {
                                'line': 18,
                                'args': ['content_by_lua_block'],
                                'block': [
                                    {
                                        'comment': ' stuff',
                                        'line': 19,
                                        'args': [],
                                        'directive': '#'
                                    }
                                ],
                                'directive': 'upstream'
                            }
                        ],
                        'directive': 'http'
                    }
                ]
            }
        ]
    }


def test_build_lua_blocks_simple(tmpdir):
    compare_parsed_and_built('lua-block-simple', 'nginx.conf', tmpdir)


def test_build_lua_blocks_larger(tmpdir):
    compare_parsed_and_built('lua-block-larger', 'nginx.conf', tmpdir)


def test_build_lua_blocks_tricky(tmpdir):
    compare_parsed_and_built('lua-block-tricky', 'nginx.conf', tmpdir)



# -*- coding: utf-8 -*-
import functools
import sys

try:
    import simplejson as json
except ImportError:
    import json

PY2 = (sys.version_info[0] == 2)
PY3 = (sys.version_info[0] == 3)

if PY2:
    input = raw_input
    basestring = basestring
else:
    input = input
    basestring = str


def fix_pep_479(generator):
    """
    Python 3.7 breaks crossplane's lexer because of PEP 479
    Read more here: https://www.python.org/dev/peps/pep-0479/
    """
    @functools.wraps(generator)
    def _wrapped_generator(*args, **kwargs):
        try:
            for x in generator(*args, **kwargs):
                yield x
        except RuntimeError:
            return

    return _wrapped_generator



# -*- coding: utf-8 -*-
from .parser import parse
from .lexer import lex
from .builder import build
from .formatter import format
from .ext.lua import LuaBlockPlugin

__all__ = ['parse', 'lex', 'build', 'format']

__title__ = 'cscrossplane'
__summary__ = 'Reliable and fast NGINX configuration file parser.'
__url__ = 'https://github.com/caas-one/crossplane'

__version__ = '0.5.4'

__author__ = 'Arie van Luttikhuizen'
__email__ = 'yulibaozi@qq.com'

__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2018 NGINX, Inc.'

default_enabled_extensions = [LuaBlockPlugin()]
for extension in default_enabled_extensions:
    extension.register_extension()



# -*- coding: utf-8 -*-
import crossplane


def test_state_directive():
    fname = '/path/to/nginx.conf'

    stmt = {
        'directive': 'state',
        'args': ['/path/to/state/file.conf'],
        'line': 5  # this is arbitrary
    }

    # the state directive should not cause errors if it's in these contexts
    good_contexts = set([
        ('http', 'upstream'),
        ('stream', 'upstream'),
        ('some_third_party_context',)
    ])

    for ctx in good_contexts:
        crossplane.analyzer.analyze(fname, stmt, term=';', ctx=ctx)

    # the state directive should not be in any of these contexts
    bad_contexts = set(crossplane.analyzer.CONTEXTS) - good_contexts

    for ctx in bad_contexts:
        try:
            crossplane.analyzer.analyze(fname, stmt, term=';', ctx=ctx)
            raise Exception("bad context for 'state' passed: " + repr(ctx))
        except crossplane.errors.NgxParserDirectiveContextError:
            continue


def test_flag_directive_args():
    fname = '/path/to/nginx.conf'
    ctx = ('events',)

    # an NGINX_CONF_FLAG directive
    stmt = {
        'directive': 'accept_mutex',
        'line': 2  # this is arbitrary
    }

    good_args = [['on'], ['off'], ['On'], ['Off'], ['ON'], ['OFF']]

    for args in good_args:
        stmt['args'] = args
        crossplane.analyzer.analyze(fname, stmt, term=';', ctx=ctx)

    bad_args = [['1'], ['0'], ['true'], ['okay'], ['']]

    for args in bad_args:
        stmt['args'] = args
        try:
            crossplane.analyzer.analyze(fname, stmt, term=';', ctx=ctx)
            raise Exception('bad args for flag directive: ' + repr(args))
        except crossplane.errors.NgxParserDirectiveArgumentsError as e:
            assert e.strerror.endswith('it must be "on" or "off"')



# -*- coding: utf-8 -*-
from .errors import NgxParserBaseException
from .builder import build
from .parser import parse


def format(filename, indent=4, tabs=False):
    payload = parse(
        filename,
        comments=True,
        single=True,
        check_ctx=False,
        check_args=False
    )

    if payload['status'] != 'ok':
        e = payload['errors'][0]
        raise NgxParserBaseException(e['error'], e['file'], e['line'])

    parsed = payload['config'][0]['parsed']
    output = build(parsed, indent=indent, tabs=tabs)
    return output



# -*- coding: utf-8 -*-
import itertools
import io

from .compat import fix_pep_479
from .errors import NgxParserSyntaxError

EXTERNAL_LEXERS = {}


@fix_pep_479
def _iterescape(iterable):
    chars = iter(iterable)
    for char in chars:
        if char == '\\':
            char = char + next(chars)
        yield char


def _iterlinecount(iterable):
    line = 1
    chars = iter(iterable)
    for char in chars:
        if char.endswith('\n'):
            line += 1
        yield (char, line)


@fix_pep_479
def _lex_file_object(file_obj):
    """
    Generates token tuples from an nginx config file object

    Yields 3-tuples like (token, lineno, quoted)
    """
    token = ''  # the token buffer
    token_line = 0  # the line the token starts on
    next_token_is_directive = True

    it = itertools.chain.from_iterable(file_obj)
    it = _iterescape(it)  # treat escaped characters differently
    it = _iterlinecount(it)  # count the number of newline characters

    for char, line in it:
        # handle whitespace
        if char.isspace():
            # if token complete yield it and reset token buffer
            if token:
                yield (token, token_line, False)
                if next_token_is_directive and token in EXTERNAL_LEXERS:
                    for custom_lexer_token in EXTERNAL_LEXERS[token](it, token):
                        yield custom_lexer_token
                        next_token_is_directive = True
                else:
                    next_token_is_directive = False
                token = ''

            # disregard until char isn't a whitespace character
            while char.isspace():
                char, line = next(it)

        # if starting comment
        if not token and char == '#':
            while not char.endswith('\n'):
                token = token + char
                char, _ = next(it)
            yield (token, line, False)
            token = ''
            continue

        if not token:
            token_line = line

        # handle parameter expansion syntax (ex: "${var[@]}")
        if token and token[-1] == '$' and char == '{':
            next_token_is_directive = False
            while token[-1] != '}' and not char.isspace():
                token += char
                char, line = next(it)

        # if a quote is found, add the whole string to the token buffer
        if char in ('"', "'"):
            # if a quote is inside a token, treat it like any other char
            if token:
                token += char
                continue

            quote = char
            char, line = next(it)
            while char != quote:
                token += quote if char == '\\' + quote else char
                char, line = next(it)

            yield (token, token_line, True)  # True because this is in quotes

            # handle quoted external directives
            if next_token_is_directive and token in EXTERNAL_LEXERS:
                for custom_lexer_token in EXTERNAL_LEXERS[token](it, token):
                    yield custom_lexer_token
                    next_token_is_directive = True
            else:
                next_token_is_directive = False

            token = ''
            continue

        # handle special characters that are treated like full tokens
        if char in ('{', '}', ';'):
            # if token complete yield it and reset token buffer
            if token:
                yield (token, token_line, False)
                token = ''

            # this character is a full token so yield it now
            yield (char, line, False)
            next_token_is_directive = True
            continue

        # append char to the token buffer
        token += char


def _balance_braces(tokens, filename=None):
    """Raises syntax errors if braces aren't balanced"""
    depth = 0

    for token, line, quoted in tokens:
        if token == '}' and not quoted:
            depth -= 1
        elif token == '{' and not quoted:
            depth += 1

        # raise error if we ever have more right braces than left
        if depth < 0:
            reason = 'unexpected "}"'
            raise NgxParserSyntaxError(reason, filename, line)
        else:
            yield (token, line, quoted)

    # raise error if we have less right braces than left at EOF
    if depth > 0:
        reason = 'unexpected end of file, expecting "}"'
        raise NgxParserSyntaxError(reason, filename, line)


def lex(filename):
    """Generates tokens from an nginx config file"""
    with io.open(filename, mode='r', encoding='utf-8') as f:
        it = _lex_file_object(f)
        it = _balance_braces(it, filename)
        for token, line, quoted in it:
            yield (token, line, quoted)


def register_external_lexer(directives, lexer):
    for directive in directives:
        EXTERNAL_LEXERS[directive] = lexer



# -*- coding: utf-8 -*-
import glob
import os
import json

from .lexer import lex
from .analyzer import analyze, enter_block_ctx
from .errors import NgxParserDirectiveError

# map of external / third-party directives to a parse function
EXTERNAL_PARSERS = {}


# TODO: raise special errors for invalid "if" args
def _prepare_if_args(stmt):
    """Removes parentheses from an "if" directive's arguments"""
    args = stmt['args']
    if args and args[0].startswith('(') and args[-1].endswith(')'):
        args[0] = args[0][1:].lstrip()
        args[-1] = args[-1][:-1].rstrip()
        start = int(not args[0])
        end = len(args) - int(not args[-1])
        args[:] = args[start:end]


def parse(filename, onerror=None, catch_errors=True, ignore=(), single=False,
        comments=False, strict=False, combine=False, check_ctx=True,
        check_args=True):
    """
    Parses an nginx config file and returns a nested dict payload

    :param filename: string contianing the name of the config file to parse
    :param onerror: function that determines what's saved in "callback"
    :param catch_errors: bool; if False, parse stops after first error
    :param ignore: list or tuple of directives to exclude from the payload
    :param combine: bool; if True, use includes to create a single config obj
    :param single: bool; if True, including from other files doesn't happen
    :param comments: bool; if True, including comments to json payload
    :param strict: bool; if True, unrecognized directives raise errors
    :param check_ctx: bool; if True, runs context analysis on directives
    :param check_args: bool; if True, runs arg count analysis on directives
    :returns: a payload that describes the parsed nginx config
    """
    config_dir = os.path.dirname(filename)

    payload = {
        'status': 'ok',
        'errors': [],
        'config': [],
    }

    # start with the main nginx config file/context
    includes = [(filename, ())]  # stores (filename, config context) tuples
    included = {filename: 0} # stores {filename: array index} map

    def _handle_error(parsing, e):
        """Adds representaions of an error to the payload"""
        file = parsing['file']
        error = str(e)
        line = getattr(e, 'lineno', None)

        parsing_error = {'error': error, 'line': line}
        payload_error = {'file': file, 'error': error, 'line': line}
        if onerror is not None:
            payload_error['callback'] = onerror(e)

        parsing['status'] = 'failed'
        parsing['errors'].append(parsing_error)

        payload['status'] = 'failed'
        payload['errors'].append(payload_error)

    def _parse(parsing, tokens, ctx=(), consume=False):
        """Recursively parses nginx config contexts"""
        fname = parsing['file']
        parsed = []

        # parse recursively by pulling from a flat stream of tokens
        for token, lineno, quoted in tokens:
            comments_in_args = []

            # we are parsing a block, so break if it's closing
            if token == '}' and not quoted:
                break

            # if we are consuming, then just continue until end of context
            if consume:
                # if we find a block inside this context, consume it too
                if token == '{' and not quoted:
                    _parse(parsing, tokens, consume=True)
                continue

            # the first token should always(?) be an nginx directive
            directive = token

            if combine:
                stmt = {
                    'file': fname,
                    'directive': directive,
                    'line': lineno,
                    'args': []
                }
            else:
                stmt = {
                    'directive': directive,
                    'line': lineno,
                    'args': []
                }

            # if token is comment
            if directive.startswith('#') and not quoted:
                if comments:
                    stmt['directive'] = '#'
                    stmt['comment'] = token[1:]
                    parsed.append(stmt)
                continue

            # TODO: add external parser checking and handling

            # parse arguments by reading tokens
            args = stmt['args']
            token, __, quoted = next(tokens)  # disregard line numbers of args
            while token not in ('{', ';', '}') or quoted:
                if token.startswith('#') and not quoted:
                    comments_in_args.append(token[1:])
                else:
                    stmt['args'].append(token)

                token, __, quoted = next(tokens)

            # consume the directive if it is ignored and move on
            if stmt['directive'] in ignore:
                # if this directive was a block consume it too
                if token == '{' and not quoted:
                    _parse(parsing, tokens, consume=True)
                continue

            # prepare arguments
            if stmt['directive'] == 'if':
                _prepare_if_args(stmt)

            try:
                # raise errors if this statement is invalid
                analyze(
                    fname=fname, stmt=stmt, term=token, ctx=ctx, strict=strict,
                    check_ctx=check_ctx, check_args=check_args
                )
            except NgxParserDirectiveError as e:
                if catch_errors:
                    _handle_error(parsing, e)

                    # if it was a block but shouldn't have been then consume
                    if e.strerror.endswith(' is not terminated by ";"'):
                        if token != '}' and not quoted:
                            _parse(parsing, tokens, consume=True)
                        else:
                            break

                    # keep on parsin'
                    continue
                else:
                    raise e

            # add "includes" to the payload if this is an include statement
            if not single and stmt['directive'] == 'include':
                pattern = args[0]
                if not os.path.isabs(args[0]):
                    pattern = os.path.join(config_dir, args[0])

                stmt['includes'] = []

                # get names of all included files
                if glob.has_magic(pattern):
                    fnames = glob.glob(pattern)
                    fnames.sort()
                else:
                    try:
                        # if the file pattern was explicit, nginx will check
                        # that the included file can be opened and read
                        open(str(pattern)).close()
                        fnames = [pattern]
                    except Exception as e:
                        fnames = []
                        e.lineno = stmt['line']
                        if catch_errors:
                            _handle_error(parsing, e)
                        else:
                            raise e

                for fname in fnames:
                    # the included set keeps files from being parsed twice
                    # TODO: handle files included from multiple contexts
                    if fname not in included:
                        included[fname] = len(includes)
                        includes.append((fname, ctx))
                    index = included[fname]
                    stmt['includes'].append(index)

            # if this statement terminated with '{' then it is a block
            if token == '{' and not quoted:
                inner = enter_block_ctx(stmt, ctx)  # get context for block
                stmt['block'] = _parse(parsing, tokens, ctx=inner)

            parsed.append(stmt)

            # add all comments found inside args after stmt is added
            for comment in comments_in_args:
                comment_stmt = {
                    'directive': '#',
                    'line': stmt['line'],
                    'args': [],
                    'comment': comment
                }
                parsed.append(comment_stmt)

        return parsed

    # the includes list grows as "include" directives are found in _parse
    for fname, ctx in includes:
        tokens = lex(fname)
        parsing = {
            'file': fname,
            'status': 'ok',
            'errors': [],
            'parsed': []
        }
        try:
            parsing['parsed'] = _parse(parsing, tokens, ctx=ctx)
        except Exception as e:
            _handle_error(parsing, e)

        payload['config'].append(parsing)

    if combine:
        return _combine_parsed_configs(payload)
    else:
        return payload


def _combine_parsed_configs(old_payload):
    """
    Combines config files into one by using include directives.

    :param old_payload: payload that's normally returned by parse()
    :return: the new combined payload
    """
    old_configs = old_payload['config']

    def _perform_includes(block):
        for stmt in block:
            if 'block' in stmt:
                stmt['block'] = list(_perform_includes(stmt['block']))
            if 'includes' in stmt:
                for index in stmt['includes']:
                    config = old_configs[index]['parsed']
                    for stmt in _perform_includes(config):
                        yield stmt
            else:
                yield stmt  # do not yield include stmt itself

    combined_config = {
        'file': old_configs[0]['file'],
        'status': 'ok',
        'errors': [],
        'parsed': []
    }

    for config in old_configs:
        combined_config['errors'] += config.get('errors', [])
        if config.get('status', 'ok') == 'failed':
            combined_config['status'] = 'failed'

    first_config = old_configs[0]['parsed']
    combined_config['parsed'] += _perform_includes(first_config)

    combined_payload = {
        'status': old_payload.get('status', 'ok'),
        'errors': old_payload.get('errors', []),
        'config': [combined_config]
    }
    return combined_payload


def register_external_parser(parser, directives):
    """
    :param parser: parser function
    :param directives: list of directive strings
    :return:
    """
    for directive in directives:
        EXTERNAL_PARSERS[directive] = parser

def parse_to_json(path):
    dict = parse(path)
    return json.dumps(dict)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import os
import sys
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from traceback import format_exception

from . import __version__
from .lexer import lex as lex_file
from .parser import parse as parse_file
from .builder import build as build_string, build_files, _enquote, DELIMITERS
from .formatter import format as format_file
from .compat import json, input


def _prompt_yes():
    try:
        return input('overwrite? (y/n [n]) ').lower().startswith('y')
    except (KeyboardInterrupt, EOFError):
        sys.exit(1)


def _dump_payload(obj, fp, indent):
    kwargs = {'indent': indent}
    if indent is None:
        kwargs['separators'] = ',', ':'
    fp.write(json.dumps(obj, **kwargs) + u'\n')


def parse(filename, out, indent=None, catch=None, tb_onerror=None, ignore='',
          single=False, comments=False, strict=False, combine=False):

    ignore = ignore.split(',') if ignore else []

    def callback(e):
        exc = sys.exc_info() + (10,)
        return ''.join(format_exception(*exc)).rstrip()

    kwargs = {
        'catch_errors': catch,
        'ignore': ignore,
        'combine': combine,
        'single': single,
        'comments': comments,
        'strict': strict
    }

    if tb_onerror:
        kwargs['onerror'] = callback

    payload = parse_file(filename, **kwargs)
    o = sys.stdout if out is None else io.open(out, 'w', encoding='utf-8')
    try:
        _dump_payload(payload, o, indent=indent)
    finally:
        o.close()


def build(filename, dirname=None, force=False, indent=4, tabs=False,
          header=True, stdout=False, verbose=False):

    if dirname is None:
        dirname = os.getcwd()

    # read the json payload from the specified file
    with open(filename, 'r') as fp:
        payload = json.load(fp)

    # find which files from the json payload will overwrite existing files
    if not force and not stdout:
        existing = []
        for config in payload['config']:
            path = config['file']
            if not os.path.isabs(path):
                path = os.path.join(dirname, path)
            if os.path.exists(path):
                existing.append(path)
        # ask the user if it's okay to overwrite existing files
        if existing:
            print('building {} would overwrite these files:'.format(filename))
            print('\n'.join(existing))
            if not _prompt_yes():
                print('not overwritten')
                return

    # if stdout is set then just print each file after another like nginx -T
    if stdout:
        for config in payload['config']:
            path = config['file']
            if not os.path.isabs(path):
                path = os.path.join(dirname, path)
            parsed = config['parsed']
            output = build_string(parsed, indent=indent, tabs=tabs, header=header)
            output = output.rstrip() + '\n'
            print('# ' + path + '\n' + output)
        return

    # build the nginx configuration file from the json payload
    build_files(payload, dirname=dirname, indent=indent, tabs=tabs, header=header)

    # if verbose print the paths of the config files that were created
    if verbose:
        for config in payload['config']:
            path = config['file']
            if not os.path.isabs(path):
                path = os.path.join(dirname, path)
            print('wrote to ' + path)


def lex(filename, out, indent=None, line_numbers=False):
    payload = list(lex_file(filename))
    if line_numbers:
        payload = [(token, lineno) for token, lineno, quoted in payload]
    else:
        payload = [token for token, lineno, quoted in payload]
    o = sys.stdout if out is None else io.open(out, 'w', encoding='utf-8')
    try:
        _dump_payload(payload, o, indent=indent)
    finally:
        o.close()


def minify(filename, out):
    payload = parse_file(
        filename,
        single=True,
        catch_errors=False,
        check_args=False,
        check_ctx=False,
        comments=False,
        strict=False
    )
    o = sys.stdout if out is None else io.open(out, 'w', encoding='utf-8')
    def write_block(block):
        for stmt in block:
            o.write(_enquote(stmt['directive']))
            if stmt['directive'] == 'if':
                o.write(u' (%s)' % ' '.join(map(_enquote, stmt['args'])))
            else:
                o.write(u' %s' % ' '.join(map(_enquote, stmt['args'])))
            if 'block' in stmt:
                o.write(u'{')
                write_block(stmt['block'])
                o.write(u'}')
            else:
                o.write(u';')
    try:
        write_block(payload['config'][0]['parsed'])
        o.write(u'\n')
    finally:
        o.close()


def format(filename, out, indent=4, tabs=False):
    output = format_file(filename, indent=indent, tabs=tabs)
    o = sys.stdout if out is None else io.open(out, 'w', encoding='utf-8')
    try:
        o.write(output + u'\n')
    finally:
        o.close()


class _SubparserHelpFormatter(RawDescriptionHelpFormatter):
    def _format_action(self, action):
        line = super(RawDescriptionHelpFormatter, self)._format_action(action)

        if action.nargs == 'A...':
            line = line.split('\n', 1)[-1]

        if line.startswith('    ') and line[4] != ' ':
            parts = filter(len, line.lstrip().partition(' '))
            line = '  ' + ' '.join(parts)

        return line


def parse_args(args=None):
    parser = ArgumentParser(
        formatter_class=_SubparserHelpFormatter,
        description='various operations for nginx config files',
        usage='%(prog)s <command> [options]'
    )
    parser.add_argument('-V', '--version', action='version', version='%(prog)s ' + __version__)
    subparsers = parser.add_subparsers(title='commands')

    def create_subparser(function, help):
        name = function.__name__
        prog = 'crossplane ' + name
        p = subparsers.add_parser(name, prog=prog, help=help, description=help)
        p.set_defaults(_subcommand=function)
        return p

    p = create_subparser(parse, 'parses a json payload for an nginx config')
    p.add_argument('filename', help='the nginx config file')
    p.add_argument('-o', '--out', type=str, help='write output to a file')
    p.add_argument('-i', '--indent', type=int, metavar='NUM', help='number of spaces to indent output')
    p.add_argument('--ignore', metavar='DIRECTIVES', default='', help='ignore directives (comma-separated)')
    p.add_argument('--no-catch', action='store_false', dest='catch', help='only collect first error in file')
    p.add_argument('--tb-onerror', action='store_true', help='include tracebacks in config errors')
    p.add_argument('--combine', action='store_true', help='use includes to create one single file')
    p.add_argument('--single-file', action='store_true', dest='single', help='do not include other config files')
    p.add_argument('--include-comments', action='store_true', dest='comments', help='include comments in json')
    p.add_argument('--strict', action='store_true', help='raise errors for unknown directives')

    p = create_subparser(build, 'builds an nginx config from a json payload')
    p.add_argument('filename', help='the file with the config payload')
    p.add_argument('-v', '--verbose', action='store_true', help='verbose output')
    p.add_argument('-d', '--dir', metavar='PATH', default=None, dest='dirname', help='the base directory to build in')
    p.add_argument('-f', '--force', action='store_true', help='overwrite existing files')
    g = p.add_mutually_exclusive_group()
    g.add_argument('-i', '--indent', type=int, metavar='NUM', help='number of spaces to indent output', default=4)
    g.add_argument('-t', '--tabs', action='store_true', help='indent with tabs instead of spaces')
    p.add_argument('--no-headers', action='store_false', dest='header', help='do not write header to configs')
    p.add_argument('--stdout', action='store_true', help='write configs to stdout instead')

    p = create_subparser(lex, 'lexes tokens from an nginx config file')
    p.add_argument('filename', help='the nginx config file')
    p.add_argument('-o', '--out', type=str, help='write output to a file')
    p.add_argument('-i', '--indent', type=int, metavar='NUM', help='number of spaces to indent output')
    p.add_argument('-n', '--line-numbers', action='store_true', help='include line numbers in json payload')

    p = create_subparser(minify, 'removes all whitespace from an nginx config')
    p.add_argument('filename', help='the nginx config file')
    p.add_argument('-o', '--out', type=str, help='write output to a file')

    p = create_subparser(format, 'formats an nginx config file')
    p.add_argument('filename', help='the nginx config file')
    p.add_argument('-o', '--out', type=str, help='write output to a file')
    g = p.add_mutually_exclusive_group()
    g.add_argument('-i', '--indent', type=int, metavar='NUM', help='number of spaces to indent output', default=4)
    g.add_argument('-t', '--tabs', action='store_true', help='indent with tabs instead of spaces')

    def help(command):
        if command not in parser._actions[-1].choices:
            parser.error('unknown command %r' % command)
        else:
            parser._actions[-1].choices[command].print_help()

    p = create_subparser(help, 'show help for commands')
    p.add_argument('command', help='command to show help for')

    parsed = parser.parse_args(args=args)

    # this addresses a bug that was added to argparse in Python 3.3
    if not parsed.__dict__:
        parser.error('too few arguments')

    return parsed


def main():
    kwargs = parse_args().__dict__
    func = kwargs.pop('_subcommand')
    func(**kwargs)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import os

import crossplane
from . import here


def test_includes_regular():
    dirname = os.path.join(here, 'configs', 'includes-regular')
    config = os.path.join(dirname, 'nginx.conf')
    payload = crossplane.parse(config)
    assert payload == {
        'status': 'failed',
        'errors': [
            {
                'file': os.path.join(dirname, 'conf.d', 'server.conf'),
                'error': '[Errno 2] No such file or directory: %r' % os.path.join(dirname, 'bar.conf'),
                'line': 5
            }
        ],
        'config': [
            {
                'file': os.path.join(dirname, 'nginx.conf'),
                'status': 'ok',
                'errors': [],
                'parsed': [
                    {
                        'directive': 'events',
                        'line': 1,
                        'args': [],
                        'block': []
                    },
                    {
                        'directive': 'http',
                        'line': 2,
                        'args': [],
                        'block': [
                            {
                                'directive': 'include',
                                'line': 3,
                                'args': ['conf.d/server.conf'],
                                'includes': [1]
                            }
                        ]
                    }
                ]
            },
            {
                'file': os.path.join(dirname, 'conf.d', 'server.conf'),
                'status': 'failed',
                'errors': [
                    {
                        'error': '[Errno 2] No such file or directory: %r' % os.path.join(dirname, 'bar.conf'),
                        'line': 5
                    }
                ],
                'parsed': [
                    {
                        'directive': 'server',
                        'line': 1,
                        'args': [],
                        'block': [
                            {
                                'directive': 'listen',
                                'line': 2,
                                'args': ['127.0.0.1:8080']
                            },
                            {
                                'directive': 'server_name',
                                'line': 3,
                                'args': ['default_server']
                            },
                            {
                                'directive': 'include',
                                'line': 4,
                                'args': ['foo.conf'],
                                'includes': [2]
                            },
                            {
                                'directive': 'include',
                                'line': 5,
                                'args': ['bar.conf'],
                                'includes': []
                            }
                        ]
                    }
                ]
            },
            {
                'file': os.path.join(dirname, 'foo.conf'),
                'status': 'ok',
                'errors': [],
                'parsed': [
                    {
                        'directive': 'location',
                        'line': 1,
                        'args': ['/foo'],
                        'block': [
                            {
                                'directive': 'return',
                                'line': 2,
                                'args': ['200', 'foo']
                            }
                        ]
                    }
                ]
            }
        ]
    }


def test_includes_globbed():
    dirname = os.path.join(here, 'configs', 'includes-globbed')
    config = os.path.join(dirname, 'nginx.conf')
    payload = crossplane.parse(config)
    assert payload == {
        'status': 'ok',
        'errors': [],
        'config': [
            {
                'file': os.path.join(dirname, 'nginx.conf'),
                'status': 'ok',
                'errors': [],
                'parsed': [
                    {
                        'directive': 'events',
                        'line': 1,
                        'args': [],
                        'block': []
                    },
                    {
                        'directive': 'include',
                        'line': 2,
                        'args': ['http.conf'],
                        'includes': [1]
                    }
                ]
            },
            {
                'file': os.path.join(dirname, 'http.conf'),
                'status': 'ok',
                'errors': [],
                'parsed': [
                    {
                        'directive': 'http',
                        'args': [],
                        'line': 1,
                        'block': [
                            {
                                'directive': 'include',
                                'line': 2,
                                'args': ['servers/*.conf'],
                                'includes': [2, 3]
                            }
                        ]
                    }
                ]
            },
            {
                'file': os.path.join(dirname, 'servers', 'server1.conf'),
                'status': 'ok',
                'errors': [],
                'parsed': [
                    {
                        'directive': 'server',
                        'args': [],
                        'line': 1,
                        'block': [
                            {
                                'directive': 'listen',
                                'args': ['8080'],
                                'line': 2
                            },
                            {
                                'directive': 'include',
                                'args': ['locations/*.conf'],
                                'line': 3,
                                'includes': [4, 5]
                            }
                        ]
                    }
                ]
            },
            {
                'file': os.path.join(dirname, 'servers', 'server2.conf'),
                'status': 'ok',
                'errors': [],
                'parsed': [
                    {
                        'directive': 'server',
                        'args': [],
                        'line': 1,
                        'block': [
                            {
                                'directive': 'listen',
                                'args': ['8081'],
                                'line': 2
                            },
                            {
                                'directive': 'include',
                                'args': ['locations/*.conf'],
                                'line': 3,
                                'includes': [4, 5]
                            }
                        ]
                    }
                ]
            },
            {
                'file': os.path.join(dirname, 'locations', 'location1.conf'),
                'status': 'ok',
                'errors': [],
                'parsed': [
                    {
                        'directive': 'location',
                        'args': ['/foo'],
                        'line': 1,
                        'block': [
                            {
                                'directive': 'return',
                                'args': ['200', 'foo'],
                                'line': 2
                            }
                        ]
                    }
                ]
            },
            {
                'file': os.path.join(dirname, 'locations', 'location2.conf'),
                'status': 'ok',
                'errors': [],
                'parsed': [
                    {
                        'directive': 'location',
                        'args': ['/bar'],
                        'line': 1,
                        'block': [
                            {
                                'directive': 'return',
                                'args': ['200', 'bar'],
                                'line': 2
                            }
                        ]
                    }
                ]
            }
        ]
    }


def test_includes_globbed_combined():
    dirname = os.path.join(here, 'configs', 'includes-globbed')
    config = os.path.join(dirname, 'nginx.conf')
    payload = crossplane.parse(config, combine=True)
    assert payload == {
        "status": "ok",
        "errors": [],
        "config": [
            {
                "file": os.path.join(dirname, "nginx.conf"),
                "status": "ok",
                "errors": [],
                "parsed": [
                    {
                        "directive": "events",
                        "args": [],
                        "file": os.path.join(dirname, "nginx.conf"),
                        "line": 1,
                        "block": []
                    },
                    {
                        "directive": "http",
                        "args": [],
                        "file": os.path.join(dirname, "http.conf"),
                        "line": 1,
                        "block": [
                            {
                                "directive": "server",
                                "args": [],
                                "file": os.path.join(dirname, "servers", "server1.conf"),
                                "line": 1,
                                "block": [
                                    {
                                        "directive": "listen",
                                        "args": ["8080"],
                                        "file": os.path.join(dirname, "servers", "server1.conf"),
                                        "line": 2
                                    },
                                    {
                                        "directive": "location",
                                        "args": ["/foo"],
                                        "file": os.path.join(dirname, "locations", "location1.conf"),
                                        "line": 1,
                                        "block": [
                                            {
                                                "directive": "return",
                                                "args": ["200", "foo"],
                                                "file": os.path.join(dirname, "locations", "location1.conf"),
                                                "line": 2
                                            }
                                        ]
                                    },
                                    {
                                        "directive": "location",
                                        "args": ["/bar"],
                                        "file": os.path.join(dirname, "locations", "location2.conf"),
                                        "line": 1,
                                        "block": [
                                            {
                                                "directive": "return",
                                                "args": ["200", "bar"],
                                                "file": os.path.join(dirname, "locations", "location2.conf"),
                                                "line": 2
                                            }
                                        ]
                                    }
                                ]
                            },
                            {
                                "directive": "server",
                                "args": [],
                                "file": os.path.join(dirname, "servers", "server2.conf"),
                                "line": 1,
                                "block": [
                                    {
                                        "directive": "listen",
                                        "args": ["8081"],
                                        "file": os.path.join(dirname, "servers", "server2.conf"),
                                        "line": 2
                                    },
                                    {
                                        "directive": "location",
                                        "args": ["/foo"],
                                        "file": os.path.join(dirname, "locations", "location1.conf"),
                                        "line": 1,
                                        "block": [
                                            {
                                                "directive": "return",
                                                "args": ["200", "foo"],
                                                "file": os.path.join(dirname, "locations", "location1.conf"),
                                                "line": 2
                                            }
                                        ]
                                    },
                                    {
                                        "directive": "location",
                                        "args": ["/bar"],
                                        "file": os.path.join(dirname, "locations", "location2.conf"),
                                        "line": 1,
                                        "block": [
                                            {
                                                "directive": "return",
                                                "args": ["200", "bar"],
                                                "file": os.path.join(dirname, "locations", "location2.conf"),
                                                "line": 2
                                            }
                                        ]
                                    }
                                ]
                            }
                        ]
                    }
                ]
            }
        ]
    }


def test_includes_single():
    dirname = os.path.join(here, 'configs', 'includes-regular')
    config = os.path.join(dirname, 'nginx.conf')
    payload = crossplane.parse(config, single=True)
    assert payload == {
        'status': 'ok',
        'errors': [],
        'config': [
            {
                'file': os.path.join(dirname, 'nginx.conf'),
                'status': 'ok',
                'errors': [],
                'parsed': [
                    {
                        'directive': 'events',
                        'line': 1,
                        'args': [],
                        'block': []
                    },
                    {
                        'directive': 'http',
                        'line': 2,
                        'args': [],
                        'block': [
                            {
                                'directive': 'include',
                                'line': 3,
                                'args': ['conf.d/server.conf']
                                # no 'includes' key
                            }
                        ]
                    }
                ]
            }
            # single config parsed
        ]
    }


def test_ignore_directives():
    dirname = os.path.join(here, 'configs', 'simple')
    config = os.path.join(dirname, 'nginx.conf')

    # check that you can ignore multiple directives
    payload = crossplane.parse(config, ignore=['listen', 'server_name'])
    assert payload == {
        "status": "ok",
        "errors": [],
        "config": [
            {
                "file": os.path.join(dirname, 'nginx.conf'),
                "status": "ok",
                "errors": [],
                "parsed": [
                    {
                        "directive": "events",
                        "line": 1,
                        "args": [],
                        "block": [
                            {
                                "directive": "worker_connections",
                                "line": 2,
                                "args": ["1024"]
                            }
                        ]
                    },
                    {
                        "directive": "http",
                        "line": 5,
                        "args": [],
                        "block": [
                            {
                                "directive": "server",
                                "line": 6,
                                "args": [],
                                "block": [
                                    {
                                        "directive": "location",
                                        "line": 9,
                                        "args": ["/"],
                                        "block": [
                                            {
                                                "directive": "return",
                                                "line": 10,
                                                "args": ["200", "foo bar baz"]
                                            }
                                        ]
                                    }
                                ]
                            }
                        ]
                    }
                ]
            }
        ]
    }

    # check that you can also ignore block directives
    payload = crossplane.parse(config, ignore=['events', 'server'])
    assert payload == {
        "status": "ok",
        "errors": [],
        "config": [
            {
                "file": os.path.join(dirname, 'nginx.conf'),
                "status": "ok",
                "errors": [],
                "parsed": [
                    {
                        "directive": "http",
                        "line": 5,
                        "args": [],
                        "block": []
                    }
                ]
            }
        ]
    }


def test_config_with_comments():
    dirname = os.path.join(here, 'configs', 'with-comments')
    config = os.path.join(dirname, 'nginx.conf')
    payload = crossplane.parse(config, comments=True)
    assert payload == {
       "errors" : [],
       "status" : "ok",
       "config" : [
          {
             "errors" : [],
             "parsed" : [
                {
                   "block" : [
                      {
                         "directive" : "worker_connections",
                         "args" : [
                            "1024"
                         ],
                         "line" : 2
                      }
                   ],
                   "line" : 1,
                   "args" : [],
                   "directive" : "events"
                },
                {
                   "line" : 4,
                   "directive": "#",
                   "args": [],
                   "comment" : "comment"
                },
                {
                   "block" : [
                      {
                         "args" : [],
                         "directive" : "server",
                         "line" : 6,
                         "block" : [
                            {
                               "args" : [
                                  "127.0.0.1:8080"
                               ],
                               "directive" : "listen",
                               "line" : 7
                            },
                            {
                               "args": [],
                               "directive": "#",
                               "comment" : "listen",
                               "line" : 7
                            },
                            {
                               "args" : [
                                  "default_server"
                               ],
                               "directive" : "server_name",
                               "line" : 8
                            },
                            {
                               "block" : [
                                  {
                                     "args": [],
                                     "directive": "#",
                                     "line" : 9,
                                     "comment" : "# this is brace"
                                  },
                                  {
                                     "args": [],
                                     "directive": "#",
                                     "line" : 10,
                                     "comment" : " location /"
                                  },
                                  {
                                     "line" : 11,
                                     "directive" : "return",
                                     "args" : [
                                        "200",
                                        "foo bar baz"
                                     ]
                                  }
                               ],
                               "line" : 9,
                               "directive" : "location",
                               "args" : [
                                  "/"
                               ]
                            }
                         ]
                      }
                   ],
                   "line" : 5,
                   "args" : [],
                   "directive" : "http"
                }
             ],
             "status" : "ok",
             "file" : config
          }
       ]
    }


def test_config_without_comments():
    dirname = os.path.join(here, 'configs', 'with-comments')
    config = os.path.join(dirname, 'nginx.conf')
    payload = crossplane.parse(config, comments=False)
    assert payload == {
       "errors" : [],
       "status" : "ok",
       "config" : [
          {
             "errors" : [],
             "parsed" : [
                {
                   "block" : [
                      {
                         "directive" : "worker_connections",
                         "args" : [
                            "1024"
                         ],
                         "line" : 2
                      }
                   ],
                   "line" : 1,
                   "args" : [],
                   "directive" : "events"
                },
                {
                   "block" : [
                      {
                         "args" : [],
                         "directive" : "server",
                         "line" : 6,
                         "block" : [
                            {
                               "args" : [
                                  "127.0.0.1:8080"
                               ],
                               "directive" : "listen",
                               "line" : 7
                            },
                            {
                               "args" : [
                                  "default_server"
                               ],
                               "directive" : "server_name",
                               "line" : 8
                            },
                            {
                               "block" : [
                                  {
                                     "line" : 11,
                                     "directive" : "return",
                                     "args" : [
                                        "200",
                                        "foo bar baz"
                                     ]
                                  }
                               ],
                               "line" : 9,
                               "directive" : "location",
                               "args" : [
                                  "/"
                               ]
                            }
                         ]
                      }
                   ],
                   "line" : 5,
                   "args" : [],
                   "directive" : "http"
                }
             ],
             "status" : "ok",
             "file" : os.path.join(dirname, 'nginx.conf')
          }
       ]
    }


def test_parse_strict():
    dirname = os.path.join(here, 'configs', 'spelling-mistake')
    config = os.path.join(dirname, 'nginx.conf')
    payload = crossplane.parse(config, comments=True, strict=True)
    assert payload == {
        'status' : 'failed',
        'errors' : [
            {
                'file': os.path.join(dirname, 'nginx.conf'),
                'error': 'unknown directive "proxy_passs" in %s:7' % os.path.join(dirname, 'nginx.conf'),
                'line': 7
            }
        ],
        'config' : [
           {
              'file' : os.path.join(dirname, 'nginx.conf'),
              'status' : 'failed',
              'errors' : [
                  {
                     'error': 'unknown directive "proxy_passs" in %s:7' % os.path.join(dirname, 'nginx.conf'),
                     'line': 7
                  }
              ],
              'parsed' : [
                 {
                    'directive' : 'events',
                    'line' : 1,
                    'args' : [],
                    'block' : []
                 },
                 {
                    'directive' : 'http',
                    'line' : 3,
                    'args' : [],
                    'block' : [
                       {
                          'directive' : 'server',
                          'line' : 4,
                          'args' : [],
                          'block' : [
                             {
                                'directive' : 'location',
                                'line' : 5,
                                'args' : ['/'],
                                'block' : [
                                   {
                                       'directive' : '#',
                                       'line' : 6,
                                       'args' : [],
                                       'comment': 'directive is misspelled'
                                   }
                                ]
                             }
                          ]
                       }
                    ]
                 }
              ]
           }
        ]
    }


def test_parse_missing_semicolon():
    dirname = os.path.join(here, 'configs', 'missing-semicolon')

    # test correct error is raised when broken proxy_pass is in upper block
    above_config = os.path.join(dirname, 'broken-above.conf')
    above_payload = crossplane.parse(above_config)
    assert above_payload == {
        "status": "failed",
        "errors": [
            {
                "file": above_config,
                "error": "directive \"proxy_pass\" is not terminated by \";\" in %s:4" % above_config,
                "line": 4
            }
        ],
        "config": [
            {
                "file": above_config,
                "status": "failed",
                "errors": [
                    {
                        "error": "directive \"proxy_pass\" is not terminated by \";\" in %s:4" % above_config,
                        "line": 4
                    }
                ],
                "parsed": [
                    {
                        "directive": "http",
                        "line": 1,
                        "args": [],
                        "block": [
                            {
                                "directive": "server",
                                "line": 2,
                                "args": [],
                                "block": [
                                    {
                                        "directive": "location",
                                        "line": 3,
                                        "args": ["/is-broken"],
                                        "block": []
                                    },
                                    {
                                        "directive": "location",
                                        "line": 6,
                                        "args": ["/not-broken"],
                                        "block": [
                                            {
                                                "directive": "proxy_pass",
                                                "line": 7,
                                                "args": ["http://not.broken.example"]
                                            }
                                        ]
                                    }
                                ]
                            }
                        ]
                    }
                ]
            }
        ]
    }

    # test correct error is raised when broken proxy_pass is in lower block
    below_config = os.path.join(dirname, 'broken-below.conf')
    below_payload = crossplane.parse(below_config)
    assert below_payload == {
        "status": "failed",
        "errors": [
            {
                "file": below_config,
                "error": "directive \"proxy_pass\" is not terminated by \";\" in %s:7" % below_config,
                "line": 7
            }
        ],
        "config": [
            {
                "file": below_config,
                "status": "failed",
                "errors": [
                    {
                        "error": "directive \"proxy_pass\" is not terminated by \";\" in %s:7" % below_config,
                        "line": 7
                    }
                ],
                "parsed": [
                    {
                        "directive": "http",
                        "line": 1,
                        "args": [],
                        "block": [
                            {
                                "directive": "server",
                                "line": 2,
                                "args": [],
                                "block": [
                                    {
                                        "directive": "location",
                                        "line": 3,
                                        "args": ["/not-broken"],
                                        "block": [
                                            {
                                                "directive": "proxy_pass",
                                                "line": 4,
                                                "args": ["http://not.broken.example"]
                                            }
                                        ]
                                    },
                                    {
                                        "directive": "location",
                                        "line": 6,
                                        "args": ["/is-broken"],
                                        "block": []
                                    }
                                ]
                            }
                        ]
                    }
                ]
            }
        ]
    }


def test_combine_parsed_missing_values():
    dirname = os.path.join(here, 'configs', 'includes-regular')
    config = os.path.join(dirname, 'nginx.conf')
    separate = {
        "config": [
            {
                "file": "example1.conf",
                "parsed": [
                    {
                        "directive": "include",
                        "line": 1,
                        "args": ["example2.conf"],
                        "includes": [1]
                    }
                ]
            },
            {
                "file": "example2.conf",
                "parsed": [
                    {
                        "directive": "events",
                        "line": 1,
                        "args": [],
                        "block": []
                    },
                    {
                        "directive": "http",
                        "line": 2,
                        "args": [],
                        "block": []
                    }
                ]
            }
        ]
    }
    combined = crossplane.parser._combine_parsed_configs(separate)
    assert combined == {
        "status": "ok",
        "errors": [],
        "config": [
            {
                "file": "example1.conf",
                "status": "ok",
                "errors": [],
                "parsed": [
                    {
                        "directive": "events",
                        "line": 1,
                        "args": [],
                        "block": []
                    },
                    {
                        "directive": "http",
                        "line": 2,
                        "args": [],
                        "block": []
                    }
                ]
            }
        ]
    }


def test_comments_between_args():
    dirname = os.path.join(here, 'configs', 'comments-between-args')
    config = os.path.join(dirname, 'nginx.conf')
    payload = crossplane.parse(config, comments=True)
    assert payload == {
        'status': 'ok',
        'errors': [],
        'config': [
            {
                'file': config,
                'status': 'ok',
                'errors': [],
                'parsed': [
                    {
                        'directive': 'http',
                        'line': 1,
                        'args': [],
                        'block': [
                            {
                                'directive': '#',
                                'line': 1,
                                'args': [],
                                'comment': 'comment 1'
                            },
                            {
                                'directive': 'log_format',
                                'line': 2,
                                'args': ['\\#arg\\ 1', '#arg 2']
                            },
                            {
                                'directive': '#',
                                'line': 2,
                                'args': [],
                                'comment': 'comment 2'
                            },
                            {
                                'directive': '#',
                                'line': 2,
                                'args': [],
                                'comment': 'comment 3'
                            },
                            {
                                'directive': '#',
                                'line': 2,
                                'args': [],
                                'comment': 'comment 4'
                            },
                            {
                                'directive': '#',
                                'line': 2,
                                'args': [],
                                'comment': 'comment 5'
                            }
                        ]
                    }
                ]
            }
        ]
    }



# -*- coding: utf-8 -*-
import os

import crossplane
from . import compare_parsed_and_built


def test_build_nested_and_multiple_args():
    payload = [
        {
            "directive": "events",
            "args": [],
            "block": [
                {
                    "directive": "worker_connections",
                    "args": ["1024"]
                }
            ]
        },
        {
            "directive": "http",
            "args": [],
            "block": [
                {
                    "directive": "server",
                    "args": [],
                    "block": [
                        {
                            "directive": "listen",
                            "args": ["127.0.0.1:8080"]
                        },
                        {
                            "directive": "server_name",
                            "args": ["default_server"]
                        },
                        {
                            "directive": "location",
                            "args": ["/"],
                            "block": [
                                {
                                    "directive": "return",
                                    "args": ["200", "foo bar baz"]
                                }
                            ]
                        }
                    ]
                }
            ]
        }
    ]
    built = crossplane.build(payload, indent=4, tabs=False)
    assert built == '\n'.join([
        'events {',
        '    worker_connections 1024;',
        '}',
        'http {',
        '    server {',
        '        listen 127.0.0.1:8080;',
        '        server_name default_server;',
        '        location / {',
        "            return 200 'foo bar baz';",
        '        }',
        '    }',
        '}'
    ])


def test_build_with_comments():
    payload = [
        {
            "directive": "events",
            "line": 1,
            "args": [],
            "block": [
                {
                    "directive": "worker_connections",
                    "line": 2,
                    "args": ["1024"]
                }
            ]
        },
        {
            "directive": "#",
            "line": 4,
            "args": [],
            "comment": "comment"
        },
        {
            "directive": "http",
            "line": 5,
            "args": [],
            "block": [
                {
                    "directive": "server",
                    "line": 6,
                    "args": [],
                    "block": [
                        {
                            "directive": "listen",
                            "line": 7,
                            "args": ["127.0.0.1:8080"]
                        },
                        {
                            "directive": "#",
                            "line": 7,
                            "args": [],
                            "comment": "listen"
                        },
                        {
                            "directive": "server_name",
                            "line": 8,
                            "args": ["default_server"]
                        },
                        {
                            "directive": "location",
                            "line": 9,
                            "args": ["/"],
                            "block": [
                                {
                                    "directive": "#",
                                    "line": 9,
                                    "args": [],
                                    "comment": "# this is brace"
                                },
                                {
                                    "directive": "#",
                                    "line": 10,
                                    "args": [],
                                    "comment": " location /"
                                },
                                {
                                    "directive": "#",
                                    "line": 11,
                                    "args": [],
                                    "comment": " is here"
                                },
                                {
                                    "directive": "return",
                                    "line": 12,
                                    "args": ["200", "foo bar baz"]
                                }
                            ]
                        }
                    ]
                }
            ]
        }
    ]
    built = crossplane.build(payload, indent=4, tabs=False)
    assert built == '\n'.join([
        'events {',
        '    worker_connections 1024;',
        '}',
        '#comment',
        'http {',
        '    server {',
        '        listen 127.0.0.1:8080; #listen',
        '        server_name default_server;',
        '        location / { ## this is brace',
        '            # location /',
        '            # is here',
        "            return 200 'foo bar baz';",
        '        }',
        '    }',
        '}'
    ])


def test_build_starts_with_comments():
    payload = [
        {
            "directive": "#",
            "line": 1,
            "args": [],
            "comment": " foo"
        },
        {
            "directive": "user",
            "line": 5,
            "args": ["root"]
        }
    ]
    built = crossplane.build(payload, indent=4, tabs=False)
    assert built == '# foo\nuser root;'


def test_build_with_quoted_unicode():
    payload = [
        {
            "directive": "env",
            "line": 1,
            "args": ["русский текст"],
        }
    ]
    built = crossplane.build(payload, indent=4, tabs=False)
    assert built == u"env 'русский текст';"


def test_build_multiple_comments_on_one_line():
    payload = [
        {
            "directive": "#",
            "line": 1,
            "args": [],
            "comment": "comment1"
        },
        {
            "directive": "user",
            "line": 2,
            "args": ["root"]
        },
        {
            "directive": "#",
            "line": 2,
            "args": [],
            "comment": "comment2"
        },
        {
            "directive": "#",
            "line": 2,
            "args": [],
            "comment": "comment3"
        }
    ]
    built = crossplane.build(payload, indent=4, tabs=False)
    assert built == '#comment1\nuser root; #comment2 #comment3'



def test_build_files_with_missing_status_and_errors(tmpdir):
    assert len(tmpdir.listdir()) == 0
    payload = {
        "config": [
            {
                "file": "nginx.conf",
                "parsed": [
                    {
                        "directive": "user",
                        "line": 1,
                        "args": ["nginx"],
                    }
                ]
            }
        ]
    }
    crossplane.builder.build_files(payload, dirname=tmpdir.strpath)
    built_files = tmpdir.listdir()
    assert len(built_files) == 1
    assert built_files[0].strpath == os.path.join(tmpdir.strpath, 'nginx.conf')
    assert built_files[0].read_text('utf-8') == 'user nginx;\n'


def test_build_files_with_unicode(tmpdir):
    assert len(tmpdir.listdir()) == 0
    payload = {
        "status": "ok",
        "errors": [],
        "config": [
            {
                "file": "nginx.conf",
                "status": "ok",
                "errors": [],
                "parsed": [
                    {
                        "directive": "user",
                        "line": 1,
                        "args": [u"測試"],
                    }
                ]
            }
        ]
    }
    crossplane.builder.build_files(payload, dirname=tmpdir.strpath)
    built_files = tmpdir.listdir()
    assert len(built_files) == 1
    assert built_files[0].strpath == os.path.join(tmpdir.strpath, 'nginx.conf')
    assert built_files[0].read_text('utf-8') == u'user 測試;\n'


def test_compare_parsed_and_built_simple(tmpdir):
    compare_parsed_and_built('simple', 'nginx.conf', tmpdir)


def test_compare_parsed_and_built_messy(tmpdir):
    compare_parsed_and_built('messy', 'nginx.conf', tmpdir)


def test_compare_parsed_and_built_messy_with_comments(tmpdir):
    compare_parsed_and_built('with-comments', 'nginx.conf', tmpdir, comments=True)


def test_compare_parsed_and_built_empty_map_values(tmpdir):
    compare_parsed_and_built('empty-value-map', 'nginx.conf', tmpdir)


def test_compare_parsed_and_built_russian_text(tmpdir):
    compare_parsed_and_built('russian-text', 'nginx.conf', tmpdir)


def test_compare_parsed_and_built_quoted_right_brace(tmpdir):
    compare_parsed_and_built('quoted-right-brace', 'nginx.conf', tmpdir)


def test_compare_parsed_and_built_directive_with_space(tmpdir):
    compare_parsed_and_built('directive-with-space', 'nginx.conf', tmpdir)



# -*- coding: utf-8 -*-
import codecs
import os
import re

from .compat import PY2

DELIMITERS = ('{', '}', ';')
EXTERNAL_BUILDERS = {}
ESCAPE_SEQUENCES_RE = re.compile(r'(\\x[0-9a-f]{2}|\\[0-7]{1,3})')


def _escape(string):
    prev, char = '', ''
    for char in string:
        if prev == '\\' or prev + char == '${':
            prev += char
            yield prev
            continue
        if prev == '$':
            yield prev
        if char not in ('\\', '$'):
            yield char
        prev = char
    if char in ('\\', '$'):
        yield char


def _needs_quotes(string):
    if string == '':
        return True

    # lexer should throw an error when variable expansion syntax
    # is messed up, but just wrap it in quotes for now I guess
    chars = _escape(string)

    # arguments can't start with variable expansion syntax
    char = next(chars)
    if char.isspace() or char in ('{', '}', ';', '"', "'", '${'):
        return True

    expanding = False
    for char in chars:
        if char.isspace() or char in ('{', ';', '"', "'"):
            return True
        elif char == ('${' if expanding else '}'):
            return True
        elif char == ('}' if expanding else '${'):
            expanding = not expanding

    return char in ('\\', '$') or expanding


def _replace_escape_sequences(match):
    return match.group(1).decode('string-escape')


def _enquote(arg):
    if not _needs_quotes(arg):
        return arg

    if PY2:
        arg = codecs.encode(arg, 'utf-8') if isinstance(arg, unicode) else arg
        arg = codecs.decode(arg, 'raw-unicode-escape')
        arg = repr(arg).replace('\\\\', '\\').lstrip('u')
        arg = ESCAPE_SEQUENCES_RE.sub(_replace_escape_sequences, arg)
        arg = unicode(arg, 'utf-8')
    else:
        arg = codecs.decode(arg, 'unicode-internal')
        arg = repr(arg).replace('\\\\', '\\')

    return arg


def build(payload, indent=4, tabs=False, header=False):
    padding = '\t' if tabs else ' ' * indent

    head = ''
    if header:
        head += '# This config was built from JSON using NGINX crossplane.\n'
        head += '# If you encounter any bugs please report them here:\n'
        head += '# https://github.com/nginxinc/crossplane/issues\n'
        head += '\n'

    def _build_block(output, block, depth, last_line):
        margin = padding * depth

        for stmt in block:
            directive = _enquote(stmt['directive'])
            line = stmt.get('line', 0)

            if directive == '#' and line == last_line:
                output += ' #' + stmt['comment']
                continue
            elif directive == '#':
                built = '#' + stmt['comment']
            elif directive in EXTERNAL_BUILDERS:
                external_builder = EXTERNAL_BUILDERS[directive]
                built = external_builder(stmt, padding, indent, tabs)
            else:
                args = [_enquote(arg) for arg in stmt['args']]

                if directive == 'if':
                    built = 'if (' + ' '.join(args) + ')'
                elif args:
                    built = directive + ' ' + ' '.join(args)
                else:
                    built = directive

                if stmt.get('block') is None:
                    built += ';'
                else:
                    built += ' {'
                    built = _build_block(built, stmt['block'], depth+1, line)
                    built += '\n' + margin + '}'

            output += ('\n' if output else '') + margin + built
            last_line = line

        return output

    body = ''
    body = _build_block(body, payload, 0, 0)
    return head + body


def build_files(payload, dirname=None, indent=4, tabs=False, header=False):
    """
    Uses a full nginx config payload (output of crossplane.parse) to build
    config files, then writes those files to disk.
    """
    if dirname is None:
        dirname = os.getcwd()

    for config in payload['config']:
        path = config['file']
        if not os.path.isabs(path):
            path = os.path.join(dirname, path)

        # make directories that need to be made for the config to be built
        dirpath = os.path.dirname(path)
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)

        # build then create the nginx config file using the json payload
        parsed = config['parsed']
        output = build(parsed, indent=indent, tabs=tabs, header=header)
        output = output.rstrip() + '\n'
        with codecs.open(path, 'w', encoding='utf-8') as fp:
            fp.write(output)


def register_external_builder(builder, directives):
    for directive in directives:
        EXTERNAL_BUILDERS[directive] = builder



# -*- coding: utf-8 -*-
import os

import crossplane
from . import here


def test_format_messy_config():
    dirname = os.path.join(here, 'configs', 'messy')
    config = os.path.join(dirname, 'nginx.conf')
    output = crossplane.format(config)
    assert output == '\n'.join([
        'user nobody;',
        r'# hello\n\\n\\\n worlddd  \#\\#\\\# dfsf\n \\n \\\n ',
        'events {',
        '    worker_connections 2048;',
        '}',
        'http { #forteen',
        '    # this is a comment',
        '    access_log off;',
        '    default_type text/plain;',
        '    error_log off;',
        '    server {',
        '        listen 8083;',
        r"""        return 200 'Ser" \' \' ver\\ \ $server_addr:\$server_port\n\nTime: $time_local\n\n';""",
        '    }',
        '    server {',
        '        listen 8080;',
        '        root /usr/share/nginx/html;',
        "        location ~ '/hello/world;' {",
        '            return 301 /status.html;',
        '        }',
        '        location /foo {',
        '        }',
        '        location /bar {',
        '        }',
        '        location /\{\;\}\ #\ ab {',
        '        } # hello',
        '        if ($request_method = P\{O\)\###\;ST) {',
        '        }',
        '        location /status.html {',
        "            try_files '/abc/${uri} /abc/${uri}.html' =404;",
        '        }',
        r"        location '/sta;\n                    tus' {",
        '            return 302 /status.html;',
        '        }',
        '        location /upstream_conf {',
        '            return 200 /status.html;',
        '        }',
        '    }',
        '    server {',
        '    }',
        '}'
    ])


def test_format_not_main_file():
    dirname = os.path.join(here, 'configs', 'includes-globbed', 'servers')
    config = os.path.join(dirname, 'server1.conf')
    output = crossplane.format(config)
    assert output == '\n'.join([
        'server {',
        '    listen 8080;',
        '    include locations/*.conf;',
        '}'
    ])


def test_format_args_not_analyzed():
    dirname = os.path.join(here, 'configs', 'bad-args')
    config = os.path.join(dirname, 'nginx.conf')
    output = crossplane.format(config)
    assert output == '\n'.join([
        'user;',
        'events {',
        '}',
        'http {',
        '}'
    ])


def test_format_with_comments():
    dirname = os.path.join(here, 'configs', 'with-comments')
    config = os.path.join(dirname, 'nginx.conf')
    output = crossplane.format(config)
    assert output == '\n'.join([
        'events {',
        '    worker_connections 1024;',
        '}',
        '#comment',
        'http {',
        '    server {',
        '        listen 127.0.0.1:8080; #listen',
        '        server_name default_server;',
        '        location / { ## this is brace',
        '            # location /',
        "            return 200 'foo bar baz';",
        '        }',
        '    }',
        '}'
    ])



# -*- coding: utf-8 -*-


class NgxParserBaseException(Exception):
    def __init__(self, strerror, filename, lineno):
        self.args = (strerror, filename, lineno)
        self.filename = filename
        self.lineno = lineno
        self.strerror = strerror

    def __str__(self):
        if self.lineno is not None:
            return '%s in %s:%s' % self.args
        else:
            return '%s in %s' % self.args


class NgxParserSyntaxError(NgxParserBaseException):
    pass


class NgxParserDirectiveError(NgxParserBaseException):
    pass


class NgxParserDirectiveArgumentsError(NgxParserDirectiveError):
    pass


class NgxParserDirectiveContextError(NgxParserDirectiveError):
    pass


class NgxParserDirectiveUnknownError(NgxParserDirectiveError):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import io
import os
import shutil
import sys

from setuptools import find_packages, setup, Command

from crossplane import (
    __title__, __summary__, __url__, __version__, __author__, __email__,
    __license__
)

here = os.path.abspath(os.path.dirname(__file__))


def get_readme():
    path = os.path.join(here, 'README.md')
    with io.open(path, encoding='utf-8') as f:
        return '\n' + f.read()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            shutil.rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        sys.exit()


setup(
    name=__title__,
    version=__version__,
    description=__summary__,
    long_description=get_readme(),
    long_description_content_type='text/markdown',
    author=__author__,
    author_email=__email__,
    url=__url__,
    packages=find_packages(exclude=['tests','tests.*']),
    license=__license__,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    entry_points={
        'console_scripts': [
            'crossplane = crossplane.__main__:main'
        ],
    },
    cmdclass={'upload': UploadCommand}
)
